/// <reference types="node" />
import { IncomingMessage, ServerResponse } from 'http';
import { EventEmitter } from 'events';
import { Store as ExpressStore } from 'express-session';
interface SessionOptions {
    genid?: any;
    name?: string;
    key?: string;
    store?: MicroStore;
    resave?: boolean;
    rolling?: boolean;
    saveUninitialized?: boolean;
    secret: string | string[];
    unset?: string;
    cookie?: Express.SessionCookieData;
}
declare class MicroCookie implements Express.SessionCookie {
    _expires: Date | boolean;
    originalMaxAge: number;
    secure: boolean;
    httpOnly: boolean;
    path: string;
    domain?: string;
    sameSite: boolean | 'lax' | 'strict' | 'none' | undefined;
    constructor(options?: Express.SessionCookieData);
    expires: Date | boolean;
    maxAge: number | null;
    readonly data: {
        originalMaxAge: number;
        expires: Date | undefined;
        secure: boolean;
        httpOnly: boolean;
        domain: string | undefined;
        path: string;
        sameSite: boolean | "none" | "lax" | "strict" | undefined;
    };
    serialize(name: string, val: string): string;
}
declare class MicroSession implements Express.Session {
    id: string;
    req: any;
    store: MicroStore;
    cookie: MicroCookie;
    [key: string]: any;
    constructor(req: any, sessionData: any);
    resetMaxAge(): this;
    touch(): this;
    private data;
    save(cb?: (err?: any) => void): this;
    reload(cb?: (err?: any) => void): this;
    destroy(cb?: (err?: any) => void): this;
    regenerate(cb: (err?: any) => void): this;
}
declare abstract class MicroStore extends EventEmitter {
    constructor();
    abstract destroy(id: string, cb?: (err?: any) => void): void;
    abstract get(id: string, cb: (err?: any, session?: any) => void): void;
    abstract set(id: string, session: any, cb?: (err?: any) => void): void;
    abstract touch(id: string, session: any, cb?: (err?: any) => void): void;
    generate?: (req: ReqAndSessionInfo) => void;
    regenerate(req: any, cb?: (err?: any) => void): void;
    load(id: string, cb: (err: any, session?: Express.SessionData | null) => any): void;
    createSession(req: any, session: any): any;
}
declare class MicroMemoryStore extends MicroStore implements ExpressStore {
    private sessions;
    constructor();
    all(cb: (err?: any, sessions?: {
        [id: string]: any;
    }) => void): void;
    clear(cb?: (err?: any) => void): void;
    destroy(id: string, cb?: (err?: any) => void): void;
    get(id: string, cb: (err?: any, session?: any) => void): void;
    set(id: string, session: any, cb?: (err?: any) => void): void;
    length(cb: (err?: any, length?: number) => void): void;
    touch(id: string, session: any, cb?: (err?: any) => void): void;
    getSession(id: string): any;
}
interface ReqAndSessionInfo extends IncomingMessage {
    sessionID: string;
    session: MicroSession;
    sessionStore: MicroStore;
}
declare function SessionManager(options?: SessionOptions): (req: IncomingMessage, res: ServerResponse) => Promise<MicroSession>;
declare namespace SessionManager {
    var Cookie: typeof MicroCookie;
    var MemoryStore: typeof MicroMemoryStore;
    var Store: typeof MicroStore;
    var Session: typeof MicroSession;
}
export = SessionManager;
