"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var uid_safe_1 = require("uid-safe");
var events_1 = require("events");
var cookie_1 = __importDefault(require("cookie"));
var url_1 = __importDefault(require("url"));
var on_headers_1 = __importDefault(require("on-headers"));
var cookie_signature_1 = __importDefault(require("cookie-signature"));
var crypto_1 = __importDefault(require("crypto"));
function generateSessionId() {
    return uid_safe_1.sync(24);
}
var MicroCookie = (function () {
    function MicroCookie(options) {
        this._expires = false;
        this.secure = false;
        this.httpOnly = true;
        this.path = '/';
        var opts = options || {};
        this.expires = opts.expires || false;
        this.path = opts.path || '/';
        this.maxAge = opts.maxAge;
        this.originalMaxAge = opts.originalMaxAge === undefined ? opts.maxAge || 0 : opts.originalMaxAge;
        this.domain = opts.domain;
        this.sameSite = opts.sameSite;
    }
    Object.defineProperty(MicroCookie.prototype, "expires", {
        get: function () {
            return this._expires;
        },
        set: function (to) {
            this._expires = to;
            if (this.maxAge) {
                this.originalMaxAge = this.maxAge;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MicroCookie.prototype, "maxAge", {
        get: function () {
            return this.expires instanceof Date ? this.expires.valueOf() - Date.now() : null;
        },
        set: function (ms) {
            this.expires = ms ? new Date(Date.now() + ms) : false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MicroCookie.prototype, "data", {
        get: function () {
            return {
                originalMaxAge: this.originalMaxAge,
                expires: this._expires instanceof Date ? this._expires : undefined,
                secure: this.secure,
                httpOnly: this.httpOnly,
                domain: this.domain,
                path: this.path,
                sameSite: this.sameSite
            };
        },
        enumerable: true,
        configurable: true
    });
    MicroCookie.prototype.serialize = function (name, val) {
        return cookie_1.default.serialize(name, val, this.data);
    };
    return MicroCookie;
}());
function hash(sess) {
    var str = JSON.stringify(sess, function (key, val) {
        if (key === 'cookie' || key == 'store') {
            return;
        }
        return val;
    });
    return crypto_1.default
        .createHash('sha1')
        .update(str, 'utf8')
        .digest('hex');
}
var MicroSession = (function () {
    function MicroSession(req, sessionData) {
        Object.defineProperty(this, 'req', {
            value: req
        });
        this.id = '';
        Object.defineProperty(this, 'id', {
            value: req.sessionID
        });
        this.store = req.sessionStore;
        this.cookie = sessionData.cookie;
        if (typeof sessionData == 'object') {
            for (var prop in sessionData) {
                if (!(prop in this)) {
                    this[prop] = sessionData[prop];
                }
            }
        }
    }
    MicroSession.prototype.resetMaxAge = function () {
        this.cookie.maxAge = this.cookie.originalMaxAge;
        return this;
    };
    MicroSession.prototype.touch = function () {
        this.resetMaxAge();
        return this;
    };
    MicroSession.prototype.data = function () {
        var out = {};
        for (var prop in this) {
            if (prop === 'cookie' || prop === 'store') {
                continue;
            }
            out[prop] = this[prop];
        }
        return out;
    };
    MicroSession.prototype.save = function (cb) {
        this.store.set(this.id, this.data(), cb || function () { });
        return this;
    };
    MicroSession.prototype.reload = function (cb) {
        var _this = this;
        this.store.get(this.id, function (err, sess) {
            if (err)
                return cb && cb(err);
            if (!sess)
                return cb && cb(new Error("failed to load session"));
            _this.store.createSession(_this.req, sess);
            cb && cb(null);
        });
        return this;
    };
    MicroSession.prototype.destroy = function (cb) {
        this.store.destroy(this.id, cb);
        return this;
    };
    MicroSession.prototype.regenerate = function (cb) {
        this.store.regenerate(this.req, cb);
        return this;
    };
    return MicroSession;
}());
var MicroStore = (function (_super) {
    __extends(MicroStore, _super);
    function MicroStore() {
        return _super.call(this) || this;
    }
    MicroStore.prototype.regenerate = function (req, cb) {
        var _this = this;
        var id = req.sessionID;
        this.destroy(id, function (err) {
            _this.generate && _this.generate(req);
            cb && cb(err);
        });
    };
    MicroStore.prototype.load = function (id, cb) {
        var _this = this;
        this.get(id, function (err, session) {
            if (err)
                return cb(err);
            if (!session)
                return cb(null);
            var req = {
                sessionID: id,
                sessionStore: _this
            };
            cb(null, _this.createSession(req, session));
        });
    };
    MicroStore.prototype.createSession = function (req, session) {
        var expires = session.cookie && session.cookie.expires;
        var originalMaxAge = session.cookie && session.cookie.originalMaxAge;
        session.cookie = new MicroCookie(session.cookie);
        if (typeof expires == 'string') {
            session.cookie.expires = new Date(expires);
        }
        session.cookie.originalMaxAge = originalMaxAge;
        req.session = new MicroSession({
            sessionStore: this,
            sessionID: session.id
        }, session);
        return req.session;
    };
    return MicroStore;
}(events_1.EventEmitter));
var MicroMemoryStore = (function (_super) {
    __extends(MicroMemoryStore, _super);
    function MicroMemoryStore() {
        var _this = _super.call(this) || this;
        _this.sessions = {};
        return _this;
    }
    MicroMemoryStore.prototype.all = function (cb) {
        var sessionIds = Object.keys(this.sessions);
        var sessions = {};
        for (var _i = 0, sessionIds_1 = sessionIds; _i < sessionIds_1.length; _i++) {
            var id = sessionIds_1[_i];
            var session = this.getSession(id);
            if (session) {
                sessions[id] = session;
            }
        }
        cb && setImmediate(function () { return cb(null, sessions); });
    };
    MicroMemoryStore.prototype.clear = function (cb) {
        this.sessions = {};
        cb && setImmediate(function () { return cb(null); });
    };
    MicroMemoryStore.prototype.destroy = function (id, cb) {
        delete this.sessions[id];
        cb && setImmediate(function () { return cb(null); });
    };
    MicroMemoryStore.prototype.get = function (id, cb) {
        var session = this.getSession(id);
        setImmediate(function () { return cb(null, session); });
    };
    MicroMemoryStore.prototype.set = function (id, session, cb) {
        this.sessions[id] = JSON.stringify(session);
        cb && setImmediate(function () { return cb(null); });
    };
    MicroMemoryStore.prototype.length = function (cb) {
        this.all(function (err, sessions) {
            if (err)
                return cb(err);
            if (!sessions)
                return cb(null, 0);
            cb(null, Object.keys(sessions).length);
        });
    };
    MicroMemoryStore.prototype.touch = function (id, session, cb) {
        var currentSession = this.getSession(id);
        if (currentSession) {
            currentSession.cookie = session.cookie;
            this.sessions[session] = JSON.stringify(currentSession);
        }
        cb && setImmediate(function () { return cb(null); });
    };
    MicroMemoryStore.prototype.getSession = function (id) {
        var sess = this.sessions[id];
        if (!sess) {
            return;
        }
        var sessJSON = JSON.parse(sess);
        if (sessJSON.cookie) {
            var expires = void 0;
            if (!(sessJSON.cookie.expires instanceof Date)) {
                expires = new Date(sessJSON.cookie.expires);
            }
            else {
                expires = sessJSON.cookie.expires;
            }
            if (expires && expires <= new Date(Date.now())) {
                delete this.sessions[id];
                return;
            }
        }
        return sessJSON;
    };
    return MicroMemoryStore;
}(MicroStore));
function unsigncookie(val, secrets) {
    for (var i = 0; i < secrets.length; i++) {
        var result = cookie_signature_1.default.unsign(val, secrets[i]);
        if (result !== false) {
            return result;
        }
    }
    return false;
}
function setcookie(res, name, val, secret, options) {
    var signed = 's:' + cookie_signature_1.default.sign(val, secret);
    var data = cookie_1.default.serialize(name, signed, options);
    var prev = res.getHeader('Set-Cookie') || [];
    var header = Array.isArray(prev) ? prev.concat(data) : [prev, data];
    res.setHeader('Set-Cookie', header);
}
function getcookie(req, name, secrets) {
    var header = req.headers.cookie;
    var raw;
    var val;
    if (header) {
        var cookies = cookie_1.default.parse(header);
        raw = cookies[name];
        if (raw) {
            if (raw.substr(0, 2) === 's:') {
                val = unsigncookie(raw.slice(2), secrets);
                if (val === false) {
                    val = undefined;
                }
            }
        }
    }
    return val;
}
function SessionManager(options) {
    var opts = options || {};
    var generateId = opts.genid || generateSessionId;
    var name = opts.name || opts.key || 'micro.sid';
    var store = opts.store || new MicroMemoryStore();
    var resaveSession = opts.resave;
    if (resaveSession === undefined) {
        resaveSession = true;
    }
    var rolling = opts.rolling;
    var cookieOptions = opts.cookie;
    var saveUninitialized = opts.saveUninitialized;
    if (saveUninitialized === undefined) {
        saveUninitialized = true;
    }
    if (opts.unset && opts.unset !== 'destroy' && opts.unset !== 'keep') {
        throw new TypeError('unset option must be either destroy or keep');
    }
    var secret;
    if (typeof opts.secret === 'string') {
        secret = [opts.secret];
    }
    else {
        secret = opts.secret;
    }
    if (!secret) {
        throw new TypeError('session requires options.secret');
    }
    if (process.env.NODE_ENV === 'production' && opts.store instanceof MicroMemoryStore) {
        console.warn("MemoryStore should not be used in production");
    }
    var storeReady = true;
    store.on('disconnect', function () { return storeReady = false; });
    store.on('connect', function () { return storeReady = false; });
    store.generate = function (req) {
        req.sessionID = generateId(req);
        req.sessionStore = store;
        req.session = new MicroSession(req, {});
        req.session.cookie = new MicroCookie(cookieOptions);
    };
    var storeImplementsTouch = typeof store.touch === 'function';
    return function (req, res) {
        return new Promise(function (resolve, reject) {
            if (!storeReady) {
                resolve();
                return;
            }
            var originalPath = url_1.default.parse(req.url).pathname || '/';
            if (cookieOptions && originalPath.indexOf(cookieOptions.path || '/') !== 0)
                return;
            var cookieId = getcookie(req, name, secret);
            var originalId;
            var originalHash;
            var savedHash;
            var session = undefined;
            var touched = false;
            function isModified(session) {
                return originalId !== session.id || originalHash !== hash(session);
            }
            function isSaved(sess) {
                return originalId === sess.id && savedHash === hash(session);
            }
            function shouldSetCookie(sessionID, session) {
                if (typeof sessionID !== 'string' || !session) {
                    return false;
                }
                return cookieId !== sessionID
                    ? saveUninitialized || isModified(session)
                    : rolling || session.cookie.expires != null && isModified(session);
            }
            function shouldSave(id, session) {
                if (typeof id !== 'string' || !session) {
                    return false;
                }
                return !saveUninitialized && cookieId !== id
                    ? isModified(session)
                    : !isSaved(session);
            }
            function shouldTouch(id, session) {
                if (typeof id !== 'string') {
                    return false;
                }
                return cookieId === id && !shouldSave(id, session);
            }
            on_headers_1.default(res, function () {
                if (session === undefined) {
                    return;
                }
                if (!shouldSetCookie(cookieId, session)) {
                    return;
                }
                if (!touched) {
                    session.touch();
                    touched = true;
                }
                setcookie(res, name, cookieId, secret[0], session.cookie.data);
            });
            var _end = res.end;
            var _write = res.write;
            var ended = false;
            res.end = function (chunk, encoding, cb) {
                if (ended) {
                    return false;
                }
                ended = true;
                var ret;
                var sync = true;
                function writeend() {
                    if (sync) {
                        ret = _end.call(res, chunk, encoding);
                        sync = false;
                        return;
                    }
                    _end.call(res, null, "");
                }
                function writetop() {
                    if (!sync) {
                        return ret;
                    }
                    if (chunk == null) {
                        ret = true;
                        return ret;
                    }
                    var contentLength = Number(res.getHeader('Content-Length'));
                    if (!isNaN(contentLength) && contentLength > 0) {
                        chunk = !Buffer.isBuffer(chunk)
                            ? Buffer.from(chunk, encoding)
                            : chunk;
                        encoding = undefined;
                        if (chunk.length !== 0) {
                            ret = _write.call(res, chunk.slice(0, chunk.length - 1), encoding);
                            chunk = chunk.slice(chunk.length - 1, chunk.length);
                            return ret;
                        }
                    }
                    ret = _write.call(res, chunk, encoding);
                    sync = false;
                    return ret;
                }
                if (!session) {
                    return _end.call(res, chunk, encoding);
                }
                if (!touched) {
                    session.touch();
                    touched = true;
                }
                if (shouldSave(cookieId, session)) {
                    session.save(function onsave(err) {
                        if (err) {
                            setImmediate(reject, err);
                            return;
                        }
                        writeend();
                    });
                    return writetop();
                }
                else if (storeImplementsTouch && shouldTouch(cookieId, session)) {
                    store.touch(cookieId, session, function ontouch(err) {
                        if (err) {
                            setImmediate(reject, err);
                        }
                        writeend();
                    });
                    return writetop();
                }
                return _end.call(res, chunk, encoding);
            };
            function generate() {
                var fakeReq = req;
                fakeReq.sessionID = cookieId;
                fakeReq.session = session;
                store.generate && store.generate(fakeReq);
                originalId = fakeReq.sessionID;
                originalHash = hash(fakeReq.session);
                session = fakeReq.session;
                cookieId = fakeReq.sessionID;
                wrapmethods(session);
            }
            function inflate(req, sess) {
                store.createSession(req, sess);
                originalId = req.sessionID;
                originalHash = hash(sess);
                if (!resaveSession) {
                    savedHash = originalHash;
                }
                wrapmethods(req.session);
                session = req.session;
            }
            function wrapmethods(sess) {
                var _reload = sess.reload;
                var _save = sess.save;
                function reload(callback) {
                    _reload.call(this, function () {
                        wrapmethods(session);
                        callback(arguments);
                    });
                }
                function save(cb) {
                    savedHash = hash(this);
                    _save.apply(this, [cb]);
                }
                Object.defineProperty(sess, 'reload', {
                    configurable: true,
                    enumerable: false,
                    value: reload,
                    writable: true
                });
                Object.defineProperty(sess, 'save', {
                    configurable: true,
                    enumerable: false,
                    value: save,
                    writable: true
                });
            }
            if (!cookieId) {
                generate();
                resolve(session);
                return;
            }
            store.get(cookieId, function (err, sess) {
                if (err && err.code !== 'ENOENT') {
                    reject(err);
                    return;
                }
                try {
                    if (err || !sess) {
                        generate();
                    }
                    else {
                        inflate(req, sess);
                    }
                }
                catch (e) {
                    reject(e);
                    return;
                }
                resolve(session);
            });
        });
    };
}
(function (SessionManager) {
    SessionManager.Cookie = MicroCookie;
    SessionManager.MemoryStore = MicroMemoryStore;
    SessionManager.Store = MicroStore;
    SessionManager.Session = MicroSession;
})(SessionManager || (SessionManager = {}));
module.exports = SessionManager;
